﻿using Sims3.SimIFace;
using System;
using Sims3.Gameplay.Actors;
using Sims3.Gameplay.EventSystem;
using Sims3.UI;
using Sims3.Gameplay.Abstracts;
using Sims3.Gameplay.DreamsAndPromises;
using Sims3.Gameplay.ActorSystems;
using Sims3.Gameplay.Utilities;
using Sims3.Gameplay.TuningValues;
using Sims3.Gameplay.Autonomy;
using Sims3.Gameplay.Interactions;
using Sims3.Gameplay.Skills;
using Sims3.Gameplay.Core;
using Sims3.SimIFace.SACS;
using Sims3.Gameplay.Roles;
using System.Collections.Generic;
using Sims3.Gameplay.Interfaces;

namespace Simoro_Workout
{
    public class Instantiator
    {
        [Tunable]
        internal static bool kInstantiator;

        static Instantiator()
        {
            World.OnWorldLoadFinishedEventHandler += new EventHandler(Instantiator.OnWorldLoadFinishedHandler);
        }

        public static void AddInteractions(Sim sim)
        {
            if (sim != null)
            {
                InteractionObjectPair iop = new InteractionObjectPair(SimoroWorkout.Singleton, sim);
                foreach (InteractionObjectPair pair in sim.Interactions)
                {
                    if (pair.Equals(iop))
                    {
                        return;
                    }
                }
                sim.AddInteraction(SimoroWorkout.Singleton);
            }
        }

        private static ListenerAction OnSimInstantiated(Event e)
        {
            Sim sim = e.TargetObject as Sim;
            AddInteractions(sim);
            return ListenerAction.Keep;
        }

        private static void OnWorldLoadFinishedHandler(Object sender, EventArgs e)
        {
            try
            {
                foreach (Sim sim in Sims3.Gameplay.Queries.GetObjects<Sim>())
                {
                    AddInteractions(sim);
                }
                EventTracker.AddListener(EventTypeId.kSimInstantiated, new ProcessEventDelegate(Instantiator.OnSimInstantiated));
            }
            catch (Exception)
            {
                SimpleMessageDialog.Show("Simoro_Workout", "Exception caught during 'OnWorldLoadFinishedHandler'. Sender: " + sender.ToString() + ", EventArgs: " + e.ToString());
            }
        }
    }

    public class SimoroWorkout : Interaction<Sim, Sim>
    {
        // Fields
        [Tunable]
        private static AthleticGameObject.Tuning kAthleticTuning = new AthleticGameObject.Tuning();
        [Tunable, TunableComment("Range: Positive floats.  Description:  The time in Sim days it takes a Sim to reach full muscular shape from the fully non-muscular shape.")]
        private static float kDaysToReachMuscularShape = 0.2f;
        [Tunable, TunableComment("Range: Motive values.  Description: Fatigue level at which a Sim will start playing fatigued workout animations.")]
        private static float kFatiguedAnimationMotiveValue = -50f;
        [Tunable, TunableComment("Range: Sim Minutes.  Description:  How often the Sim changes workout animations on the bench.")]
        private static float kWorkoutAnimationSwitchTime = 10f;
		[Tunable, TunableComment("Range:  Percentage chance under 1.0f.  Description:  Chance that a Sim idles when transitioning between workout interactions.")]
		public static float kChanceToIdle = 0.1f;
        
        private int mAnimLoopCounter;
        private Athletic mAthleticSkill;
        private Stage mFatigueStage;
        private AlarmHandle mPumpedAlarm;
        private SkillStage mSkillStage;
        private AlarmHandle mStagesAlarm;
        private Sim.SwitchOutfitHelper mSwitchOutfitHelper;
        private DateAndTime mWorkoutStartTime;
        public static readonly InteractionDefinition Singleton = new SimoroWorkout.Definition();

        // Methods
        protected override void AddExcludedDreams(ICollection<DreamNames> excludedDreams)
        {
            base.AddExcludedDreams(excludedDreams);
            excludedDreams.Add(DreamNames.workout);
            excludedDreams.Add(DreamNames.do_a_strength_workout);
        }

        protected void AddPumped()
        {
            if (this.mAthleticSkill.SkillLevel >= SimoroWorkout.kAthleticTuning.AthleticLevelToAddPumped)
            {
                this.Actor.BuffManager.AddElementPaused(BuffNames.Pumped, Origin.FromExercise);
            }
        }

        public void CalculateAthleticStages()
        {
            if (this.Actor == null || this.Actor.HasBeenDestroyed || (this.Actor.InteractionQueue == null || !this.Actor.InteractionQueue.HasInteraction(this)))
            {
                this.Actor.RemoveAlarm(this.mStagesAlarm);
            }
            else
            {
                this.DisplayAthleticStages();
            }
        }

        public override void Cleanup()
        {
            if (this.mWorkoutStartTime.Ticks != 0L)
            {
                this.WorkoutFinished();
            }
            AlarmManager.Global.RemoveAlarm(this.mPumpedAlarm);
            if (this.mSwitchOutfitHelper != null)
            {
                this.mSwitchOutfitHelper.Dispose();
                this.mSwitchOutfitHelper = null;
            }
            if (((this.Actor != null) && !this.Actor.HasBeenDestroyed) && this.StandardEntryCalled)
            {
                this.Actor.BuffManager.UnpauseBuff(BuffNames.Pumped);
                this.Actor.SimRoutingComponent.EnableDynamicFootprint();
                this.Actor.LookAtManager.EnableLookAts();
            }
            base.Cleanup();
        }

        protected virtual Stage ConfigureFatigueStage(Sim actor, Athletic athleticSkill, InteractionInstance inst)
        {
            if (athleticSkill != null)
            {
                if (athleticSkill.IsBodyBuilder())
                {
                    return null;
                }
            }
            float num = 1f - SimoroWorkout.kAthleticTuning.DelayFatiguePercentagePerAthleticSkillLevel * (float)athleticSkill.SkillLevel;
            if (num <= 0f)
            {
                num = 0f;
            }
            if (actor.TraitManager.HasElement(TraitNames.Athletic))
            {
                num *= TraitTuning.AthleticTraitSimFatigueMultiplier;
            }
            float expectedRateOfChange = inst.GetCommodityUpdate(CommodityKind.Fatigue) * num;
            return new BuffStage(SimoroWorkout.GetFatigueStageName(actor.IsFemale), CommodityKind.Fatigue, BuffNames.Fatigued, DeltaStage.Direction.Negative, true, actor, expectedRateOfChange, true, true);
        }
        
        public override void ConfigureInteraction()
        {
            Athletic.HygienePauseCallback hygienePauseCallback = new Athletic.HygienePauseCallback(this.HygienePauseFunction);
            Athletic.HygieneResumeCallback hygieneResumeCallback = new Athletic.HygieneResumeCallback(this.HygieneResumeFunction);
            this.SetAvailableTones(new List<Tone>
		    {
			    new Athletic.GoodPacingTone(),
			    new Athletic.NoSweatTone(hygienePauseCallback, hygieneResumeCallback),
			    new Athletic.PushSelfTone(),
			    new Athletic.QuickBurstTone()
		    });
            if (this.mAthleticSkill == null)
            {
                this.mAthleticSkill = (this.Actor.SkillManager.AddElement(SkillNames.Athletic) as Athletic);
            }
            if (this.Actor.SimDescription.TeenOrAbove)
            {
                this.mSkillStage = new SkillStage(LocalizeString(this.Actor.IsFemale, "SkillStage", new object[0]), SkillNames.Athletic, this.Actor, base.GetCommodityUpdate(CommodityKind.SkillAthletic), true, true);
            }
            if (!this.Actor.SimDescription.IsVampire)
            {
                this.mFatigueStage = this.ConfigureFatigueStage(this.Actor, this.mAthleticSkill, this);
            }
            this.DisplayAthleticStages();
            this.mStagesAlarm = this.Target.AddAlarmRepeating(1f, TimeUnit.Minutes, new AlarmTimerCallback(this.CalculateAthleticStages), 1f, TimeUnit.Minutes, "Work Out: Athletic Game Object", AlarmType.DeleteOnReset);
        }

        public void DisplayAthleticStages()
        {
            Stage activeStage = this.ActiveStage;
            List<Stage> list;
            if (this.mSkillStage != null)
            {
                this.mSkillStage.UpdateTimeOfCompletion(this.mSkillStage.ExpectedRateOfChange, this.Actor);
            }
            if (this.mFatigueStage != null)
            {
                float expectedRateOfChange = this.GetCommodityUpdate(CommodityKind.Fatigue) * this.GetFatigueMultiplier();
                if (this.CurrentTone is Athletic.GoodPacingTone)
                {
                    expectedRateOfChange *= Athletic.GoodPacingFatigueMultiplier;
                }
                else if (this.CurrentTone is Athletic.QuickBurstTone)
                {
                    expectedRateOfChange += Athletic.QuickBurstFatigueDelta;
                }
                if (expectedRateOfChange != (this.mFatigueStage as BuffStage).ExpectedRateOfChange)
                {
                    this.mFatigueStage = new BuffStage(SimoroWorkout.GetFatigueStageName(this.Actor.IsFemale), CommodityKind.Fatigue, BuffNames.Fatigued, DeltaStage.Direction.Negative, true, this.Actor, expectedRateOfChange, true, true);
                }
            }
            if (this.mFatigueStage != null)
            {
                list = new List<Stage>(new Stage[] { this.mFatigueStage });
                if (this.mAthleticSkill.SkillLevel < this.mAthleticSkill.NonPersistableData.MaxSkillLevel && this.mSkillStage != null && this.mSkillStage.TimeRemaining(this) < this.mFatigueStage.TimeRemaining(this))
                {
                    list.Add(this.mSkillStage);
                }
            }
            else
            {
                list = new List<Stage>(new Stage[] { new IndefiniteStage(this.GetInteractionName(), true) });
                if (this.mAthleticSkill.SkillLevel < this.mAthleticSkill.NonPersistableData.MaxSkillLevel && this.mSkillStage != null)
                {
                    list.Add(this.mSkillStage);
                }
            }
            this.Stages = list;
            if (!list.Contains(activeStage))
            {
                return;
            }
            this.ActiveStage = activeStage;      
        }

        public StateMachineClient EnterStateMachineAndInitParameters(Sim actor, bool needClothingChange)
        {
            StateMachineClient client = StateMachineClient.Acquire(actor, this.StateMachineName, AnimationPriority.kAPDefault);
            client.SetActor("x", actor);
            client.EnterState("x", "Enter");
            client.SetParameter("IsCurrentOutfitAthletic", !needClothingChange);
            client.SetParameter("skill", actor.SkillManager.GetSkillLevelParameterForJazzGraph(SkillNames.Athletic));
            client.SetParameter("fatigued", actor.Motives.GetValue(CommodityKind.Fatigue) <= SimoroWorkout.kFatiguedAnimationMotiveValue);
            return client;
        }

        public float GetFatigueMultiplier()
        {
            float num;
            if (this.mAthleticSkill.IsBodyBuilder())
            {
                num = 0.0f;
            }
            else
            {
                num = (1.0f - SimoroWorkout.kAthleticTuning.DelayFatiguePercentagePerAthleticSkillLevel * this.mAthleticSkill.SkillLevel);
                if (num <= 0.0f)
                {
                    num = 0.0f;
                }
            }
            if (this.Actor.TraitManager.HasElement(TraitNames.Athletic))
            {
                num *= TraitTuning.AthleticTraitSimFatigueMultiplier;
            }
            return num;
        }

        public static string GetFatigueStageName(bool isFemale)
        {
            return SimoroWorkout.LocalizeString(isFemale, "WorkOutFatigueStage", new object[0]);
        }

        protected void HygienePauseFunction()
        {
            base.DisableMotiveDelta(CommodityKind.Hygiene);
        }

        protected void HygieneResumeFunction()
        {
            base.EnableMotiveDelta(CommodityKind.Hygiene);
        }

        public string JazzStateName(Sim actor, InteractionDefinition definition)
        {
            return "WorkOut";
        }

        public static string LocalizeString(bool isFemale, string name, params object[] parameters)
        {
            return Localization.LocalizeString(isFemale, "Simoro_WorkOut/SimoroDummyObject/SimoroWorkOut:" + name, parameters);
        }

        public void RequestWorkOutExit(StateMachineClient smc)
        {
            smc.RequestState("x", "Stop");
        }

        protected override bool Run()
        {
            this.mSwitchOutfitHelper = new Sim.SwitchOutfitHelper(this.Actor, Sim.ClothesChangeReason.GoingToWorkOut);
            this.mSwitchOutfitHelper.Start();
            base.StandardEntry();
            this.mSwitchOutfitHelper.Wait(true);
            this.mCurrentStateMachine = this.EnterStateMachineAndInitParameters(this.Actor, this.mSwitchOutfitHelper.WillChange);
            this.Actor.RemoveAlarm(this.mStagesAlarm);
            this.mSwitchOutfitHelper.AddScriptEventHandler(this);
            if (this.mAthleticSkill == null)
            {
                this.mAthleticSkill = (this.Actor.SkillManager.AddElement(SkillNames.Athletic) as Athletic);
            }
            this.SetMaterialState(this.Actor);
            this.mCurrentStateMachine.RequestState("x", this.JazzStateName(this.Actor, base.InteractionDefinition));
            this.mPumpedAlarm = AlarmManager.Global.AddAlarm(SimoroWorkout.kAthleticTuning.MinsToAddPumped, TimeUnit.Minutes, new AlarmTimerCallback(this.AddPumped), "AthleticGameObject: Pumped", AlarmType.DeleteOnReset, this.Actor);
            float fatigueMultiplier = this.GetFatigueMultiplier();
            float multiplier = 1f;
            if (this.Actor.TraitManager.HasElement(TraitNames.Athletic))
            {
                multiplier = SimoroWorkout.kAthleticTuning.AthleticTraitDestressMultiplier;
                if (SimoroWorkout.kAthleticTuning.CanTriggerAthleticTraitTip)
                {
                    TraitTipsManager.ShowTraitTip(13271263770231521664uL, this.Actor);
                }
            }
            base.BeginCommodityUpdate(CommodityKind.Fun, multiplier);
            base.BeginCommodityUpdate(CommodityKind.Fatigue, fatigueMultiplier);
            base.BeginCommodityUpdates();
            if (base.CurrentTone is Athletic.NoSweatTone)
            {
                base.DisableMotiveDelta(CommodityKind.Hygiene);
            }
            base.StartStages();
            BuffStage buffStage = this.FatigueStage as BuffStage;
            if (buffStage != null)
            {
                buffStage.ResetEstimatedRateOfChange(base.GetCommodityUpdate(CommodityKind.Fatigue), this.Actor);
            }
            SkillStage skillStage = this.SkillStage as SkillStage;
            if (skillStage != null)
            {
                skillStage.UpdateTimeOfCompletion(base.GetCommodityUpdate(CommodityKind.SkillAthletic), this.Actor);
            }
            this.mWorkoutStartTime = SimClock.CurrentTime();
            this.StartFitness(this);
            bool flag2 = this.DoLoop(ExitReason.Default, new Interaction<Sim, Sim>.InsideLoopFunction(this.WorkOutLoop), this.mCurrentStateMachine);
            this.WorkoutFinished();
            this.RequestWorkOutExit(this.mCurrentStateMachine);
            base.EndCommodityUpdates(flag2);
            base.StandardExit(true);
            return flag2;
        }

        protected virtual void SetMaterialState(Sim actor)
        {
        }

        protected void StartFitness(InteractionInstance instance)
        {
            this.WorkoutAnimationWithoutIdle(instance.mCurrentStateMachine);
            Sim instanceActor = instance.InstanceActor;
            instanceActor.LookAtManager.DisableLookAts();
            instanceActor.AddMuscleDelta(SimoroWorkout.kDaysToReachMuscularShape);
        }

        protected void StopFitness(InteractionInstance instance, float hoursPassed)
        {
            Sim instanceActor = instance.InstanceActor;
            instanceActor.LookAtManager.EnableLookAts();
            instanceActor.ResetShapeDelta();
            Athletic skill = instanceActor.SkillManager.GetSkill<Athletic>(SkillNames.Athletic);
            if (skill != null)
            {
                skill.AddStrengthTime(hoursPassed);
            }
        }

        public float SwitchAnimationLoopTime()
        {
            return SimoroWorkout.kWorkoutAnimationSwitchTime;
        }

        public void SwitchWorkoutAnimation(StateMachineClient currentStateMachine)
        {
            if (currentStateMachine != null)
            {
                if (RandomUtil.RandomChance01(kChanceToIdle))
                {
                    currentStateMachine.RequestState(false, "x", "Idle", DriverRequestFlags.kRequestAllowRequestForCurrentState);
                }
                else
                {
                    this.WorkoutAnimationWithoutIdle(currentStateMachine);
                }
            }
        }

        public void WorkoutAnimationWithoutIdle(StateMachineClient currentStateMachine)
        {
            if (currentStateMachine != null)
            {
                switch (RandomUtil.GetInt(1))
                {
                    case 0:
                        currentStateMachine.RequestState(false, "x", "Pushups", DriverRequestFlags.kRequestAllowRequestForCurrentState);
                        break;

                    case 1:
                        currentStateMachine.RequestState(false, "x", "Crunches", DriverRequestFlags.kRequestAllowRequestForCurrentState);
                        break;
                }
            }
        }

        private void WorkoutFinished()
        {
            float num = SimClock.ElapsedTime(TimeUnit.Hours, this.mWorkoutStartTime);
            this.mWorkoutStartTime.Ticks = 0L;
            this.StopFitness(this, num);
            EventTracker.SendEvent(new IncrementalEvent(EventTypeId.kStoppedWorkingOut, this.Actor, this.Target, num));
        }

        public void WorkOutLoopParameters(StateMachineClient smc, Sim actor)
        {
            smc.SetParameter("skill", actor.SkillManager.GetSkillLevelParameterForJazzGraph(SkillNames.Athletic));
            smc.SetParameter("fatigued", actor.Motives.GetValue(CommodityKind.Fatigue) <= SimoroWorkout.kFatiguedAnimationMotiveValue);
        }

        public void WorkOutLoop(StateMachineClient smc, Interaction<Sim, Sim>.LoopData ld)
        {
            if (this.Actor.BuffManager.HasElement(BuffNames.Fatigued))
            {
                this.Actor.AddExitReason(ExitReason.Finished);
            }
            if ((ld.mLifeTime / this.SwitchAnimationLoopTime()) > (float)this.mAnimLoopCounter)
            {
                this.mAnimLoopCounter++;
                this.SwitchWorkoutAnimation(this.mCurrentStateMachine);
            }
            this.WorkOutLoopParameters(smc, this.Actor);
            EventTracker.SendEvent(new TimePassedEvent(EventTypeId.kWorkedOut, this.Actor, this.Target, ld.mDeltaTime));
            EventTracker.SendEvent(new TimePassedEvent(EventTypeId.kStrengthTimeAdded, this.Actor, this.Target, ld.mDeltaTime));
        }

        // Properties
        public override string DefaultToneDescription
        {
            get
            {
                return Localization.LocalizeString("Gameplay/Skills/Athletic:NeutralToneDescription", new object[0]);
            }
        }

        public override string DefaultToneText
        {
            get
            {
                return Localization.LocalizeString("Gameplay/Skills/Athletic:NeutralTone", new object[0]);
            }
        }

        public Stage FatigueStage
        {
            get
            {
                return this.mFatigueStage;
            }
        }

        public Stage SkillStage
        {
            get
            {
                return this.mSkillStage;
            }
        }

        public string StateMachineName
        {
            get
            {
                return "Simoro_Workout";
            }
        }

        //Nested Types
        public class Definition : InteractionDefinition<Sim, Sim, SimoroWorkout>, IStrengthInteractionDefinition
        {
            // Methods
            private string CouchPotatoTraitGreyedTooltip()
            {
                return LocalizeString("CouchPotatoTraitGreyedTooltip", new object[0]);
            }

            protected override string GetInteractionName(Sim actor, Sim target, InteractionObjectPair iop)
            {
                if (actor.HasTrait(TraitNames.Daredevil))
                {
                    return LocalizeString("DaredevilInteractionName", new object[0]);
                }
                return LocalizeString("SimoroWorkout", new object[0]);
            }

            private static string LocalizeString(string name, params object[] parameters)
            {
                return Localization.LocalizeString("Simoro_WorkOut/SimoroDummyObject:" + name, parameters);
            }

            protected override bool Test(Sim actor, Sim target, bool isAutonomous, ref GreyedOutTooltipCallback greyedOutTooltipCallback)
            {
                if (actor != target)
                {
                    return false;
                }
                if (!actor.SimDescription.TeenOrAbove)
                {
                    return false;
                }
                if (RolePaparazzi.IsSimPaparazzi(actor))
                {
                    return false;
                }
                if (actor.BuffManager.HasElement(BuffNames.Fatigued))
                {
                    greyedOutTooltipCallback = Interaction<Sim, Sim>.CreateTooltipCallback(LocalizeString("ToFatigued", new object[0]));
                    return false;
                }
                if (!actor.HasTrait(TraitNames.Athletic))
                {
                    if (!actor.IsOnHomeLot(actor))
                    {
                        return false;
                    }
                    if (actor.TraitManager.HasElement(TraitNames.CouchPotato) && actor.MoodManager.MoodValue < TraitTuning.CouchPotatoMoodForWorkOut)
                    {
                        greyedOutTooltipCallback = new GreyedOutTooltipCallback(this.CouchPotatoTraitGreyedTooltip);
                        return false;
                    }
                }
                float radius = 1.75f;
                foreach (GameObject obj in Sims3.Gameplay.Queries.GetObjects<GameObject>(actor.Position, radius))
                {
                    BlockedStatus testblocked = Route.IsRouteToObjectUnobstructed(actor.ObjectId, obj.Position, actor.SimRoutingComponent.GetRouteOptions(), actor.ObjectId, actor.SimDescription.AgeGenderSpecies);
                    if (testblocked != BlockedStatus.Unobstructed)
                    {
                        greyedOutTooltipCallback = InteractionInstance.CreateTooltipCallback(LocalizeString("NotEnoughRoom", new object[0]));
                        return false;
                    }
                }
                return true;
            }
        }
    }    
}
